/*
 * $Id: slots.c,v 1.2 1997/10/26 19:25:24 owara Exp owara $
 *
 * slots.c --
 *	A slot machine game.
 *	Fun for all ...
 *	Asks the user for number of games to play,
 *	then runs games and tallies up wins/losses.
 *
 * Name:	Shane Owara [solution set]
 * Mail:	owara@fas.harvard.edu
 * Asst:	Four -- Randomness
 */

/*
 * NOTES ON MODIFICATION:
 *	This program can be easily modified to take
 *	different costs, wheels, picture types, and prizes.
 *
 *	In general, where you see the line:
 *	 * MODIFY THIS TO CHANGE THE MACHINE
 *	there is a variable or constant that can be changed
 *	to affect the various parts of the machine.
 *
 *	General slot machine information:
 *
 *		Cost is defined by the constant COST_PER_PULL.
 *		Wheels are determined by the constant NUM_WHEELS.
 *		(Note that there must be at least two wheels.)
 *		Pictures are defined by NUM_PICTURES.
 *		(Note that this number must be based on the chances[] array.)
 *		You can find all of these #define'd at the top level.
 *
 *	Pictures and prizes are determined by several things:
 *
 *		item_t -- enum type indexing all items
 *		Found at the top level.
 *
 *		chances -- number of each picture on wheel
 *		In function spin_wheel().
 *		(Note that every time this changes, so should NUM_PICTURES.)
 *
 *		prizes -- values of winnings on pictures
 *		win_msg -- strings of win messages
 *		In function cash_out().
 *
 *	One should modify all of these in tandem
 *	to introduce a new item type or delete a current one.
 *	Record order in each field is important to maintain.
 *	Modifying a particular stat is accomplished
 *	by editing the appropriate field.
 *
 *	Technical note:  we might want to make the local statistics
 *	arrays static within the functions to keep them in memory
 *	(and not just pushed onto the functions' stacks all the time)
 *	or use structs to contain related information on pictures.
 *	But both of these techniques are beyond the current scope
 *	of the assignment.
 */

/* Header files. */
#include <stdio.h>
#include "genlib.h"
#include "simpio.h"
#include "random.h"

/*
 * Constants:
 * MODIFY THIS TO CHANGE THE MACHINE
 *	COST_PER_PULL is the price (in pennies) to play.
 *	NUM_WHEELS is the number of spinning wheels on the machine.
 *	NUM_PICTURES is the number of total pictures in the wheel.
 *	See below for constants created by an enumerated type.
 *
 *	NUM_PICTURES must not be in conflict with the chances[]
 *	array found in the spin_wheel function.
 *	Perhaps a dependent variable should be worked in here,
 *	but that would add run-time overhead and perhaps some
 *	unecessary bookkeeping.
 */
#define COST_PER_PULL    25
#define NUM_WHEELS        3
#define NUM_PICTURES     12

/*
 * New Type: item_t --
 * MODIFY THIS TO CHANGE THE MACHINE
 *	This type keeps track of the kinds of pictures on slot machines
 *	in an abstracted fashion.
 *	Note that it might be more aesthetically satisfying
 *	if the order of the type's elements were different
 *	(in descending prize/ascending chance order),
 *	but this order was specified by the graphics program.
 *	This order is maintained in the separate arrays of information
 *	found in other functions.
 *
 *	Also creates a constant with the number of kinds.
 *	This constant, NO_ITEM, is reliant on the fact that
 *	we don't care what the values of the elements are;
 *	we just need identifying tags.
 *	If the tags are initialized oddly,
 *	the count might be off ... otherwise,
 *	this is a cute way to keep track (automatically)
 *	of the number of items we have.
 *
 *	Note: the enumerated names are unused (except for NO_ITEM)
 *	by all local functions -- the names are specified for the
 *	graphics part of the assignment.
 */
typedef enum {
	BEANIE,
	SACK,
	TIE,
	CANDY,
	NO_ITEM
} item_t;

/* Function prototypes. */
int get_num_pulls(void);
void generate_pulls(int n, int win_stats[]);
void cash_out(int num_pulls, int wins[]);
item_t spin_wheel (void);

/*
 * main --
 *	A simple set of function calls that implements
 *	the slot machine pulls.
 */
int
main () {
	/*
	 * Variables:
	 * The only significant variables here are ones that carry over
	 * from function to function.
	 * num_pulls -- the number of pulls a user wants
	 * wins -- an array of the wins on each picture
	 */
	int i;
	int num_pulls;
	int wins[NO_ITEM];

	/*
	 * Initialize variables:
	 *	seed random number generator
	 *	wins are all at zero
	 */
	Randomize();
	for (i = 0; i < NO_ITEM; i++)
		wins[i] = 0;

	num_pulls = get_num_pulls();
	generate_pulls(num_pulls, wins);
	cash_out(num_pulls, wins);
	exit(0);
}

/*
 * get_num_pulls --
 *	A simple reprompt loop to get the number of pulls from user.
 *
 *	Returns the number of pulls user wants (positive).
 */
int
get_num_pulls () {
	int num_pulls;

	do {
		printf("How many pulls would you like? ");
		num_pulls = GetInteger();
		if (num_pulls <= 0)
			printf("Sorry.  Iterations must be greater than 0.\n");
	} while (num_pulls <= 0);

	return num_pulls;
}

/*
 * generate_pulls --
 *	Runs the simulated slot machine for n rounds of play,
 *	keeping track of wins for each kind of picture in an array.
 *
 *	Returns nothing, but we take advantage of the fact that
 *	arrays can be modified within functions to "return"
 *	the number of wins on each case.
 *	The array "win_stats" holds this information.
 */
void
generate_pulls (int n, int win_stats[]) {
	int i, j;
	item_t wheels[NUM_WHEELS];

	/* Simulate playing rounds, only tracking the wins. */
	for (i = 0; i < n; i++) {
		/* Spin the wheels ... */
		for (j = 0; j < NUM_WHEELS; j++)
			wheels[j] = spin_wheel();

		/* Check for a win -- all wheels must match. */
		for (j = 1; j < NUM_WHEELS; j++) {
			if (wheels[0] != wheels[j])
				break;
			else if (j == NUM_WHEELS - 1)
				win_stats[wheels[0]]++;
		}
	}
}

/*
 * cash_out --
 *	Now print out statistics, calculating purse:
 *		start purse at cost of games
 *		start pulls_lost at all games
 *		go through winnings for each item
 *		calculate losses on num_pulls
 *		print out purse loss
 *
 *	Note: There is a magic constant (100) in dollar/cent ratio
 *	used because it's a common way to break up decimals.
 *
 *	Note: "s" trick on win_msg will not work
 *	if the plural of item type is not made by adding 's'.
 *	We could change the win_msg array to a 2D,
 *	with a plural version in the second column,
 *	but that would be a bit far.
 */
void
cash_out (int num_pulls, int wins[]) {
	int i, purse, pulls_lost;

	/*
	 * MODIFY THIS TO CHANGE THE MACHINE
	 * prizes -- values for all prizes, in cents
	 * win_msg -- the winning type string messages
	 * We keep these arrays here with the function
	 * that accesses them.
	 */
	int prizes[NO_ITEM] = {
		10000,	/* BEANIE prize  $100.00 */
		  250,	/* SACK prize      $2.50 */
		 1000,	/* TIE prize      $10.00 */
		   50   /* CANDY prize     $0.50 */
	};
	string win_msg[NO_ITEM] = {
		"beanie jackpot",
		"paper bag",
		"dilbert tie",
		"candy bar"
	};

	purse = 0;
	purse -= num_pulls * COST_PER_PULL;
	pulls_lost = num_pulls;
	for (i = 0; i < NO_ITEM; i++) {
		printf("You won %d %s%s at $%d.%.2d.\n",
			wins[i],
			win_msg[i],
			wins[i] == 1 ? "" : "s",
			prizes[i] / 100,      /* prints dollars from pennies */
			prizes[i] % 100);	/* prints cents from pennies */
		purse += wins[i] * prizes[i];
		pulls_lost -= wins[i];
	}

	/* Print out lost pulls (with trick on "1 pull"). */
	printf("You lost on %d pull%s.\n\n", pulls_lost,
		pulls_lost == 1 ? "" : "s");
	
	/* Print out winnings/losses */
	if (purse == 0)
		printf("You broke even!\n\n");
	else if (purse > 0)
		printf("Congratulations! You won $%d.%.2d!!\n\n",
			purse / 100,
			purse % 100);
	else {
		/*
		 * This is a hack to get the printings of
		 * the statistics to be positive.
		 * Obviously, if the purse were to be used
		 * later in this function, I would have to
		 * do something else.
		 */
		purse *= (-1);
		printf("Sorry, but you lost $%d.%.2d.\n\n",
			purse / 100,
			purse % 100);
	}
}

/*
 * spin_wheel --
 *	Simulates the spinning of one wheel.
 *	Takes no arguments and returns the picture spun to.
 *	Statistics are based on the chances array defined below.
 *	A random number is picked as the "spin value".
 *	This value is walked down the chances array,
 *	and the appropriate value is returned when found.
 */
item_t
spin_wheel () {
	int i;
	int spin_val;

	/*
	 * MODIFY THIS TO CHANGE THE MACHINE
	 * chances -- number of pictures for each type
	 * We keep this array here with the function
	 * that accesses it.
	 */
	int chances[NO_ITEM] = {
		1,	/* number of BEANIE pictures */
		3,	/* number of SACK pictures */
		2,	/* number of TIE pictures */
		6	/* number of CANDY pictures */
	};

	spin_val = RandomInteger(1, NUM_PICTURES);
	for (i = 0; i < NO_ITEM; i++) {
		spin_val -= chances[i];
		if (spin_val <= 0)
			return (item_t)i;
	}

	/*
	 * Should never reach this part of the program.
	 * This will only happen if NUM_PICTURES is defined wrong.
	 * As mentioned eariler, we might want to get rid of this
	 * in a later version; replacing it with something calculated
	 * off of the chances[] array at initialization
	 * (if we added that now, then the value would be calculated
	 * every time we spun a wheel ... and we don't want that.
	 */
	printf("Error: spin_wheel returns invalid picture index!\n");
	exit(1);
}
